#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""!
Author: J.A. de Jong - ASCEE

Description:

Data Acquistiion (DAQ) device descriptors, and the DAQ devices themselves

"""
from dataclasses import dataclass, field
from .lasp_daqdevice import query_devices, DeviceInfo
__all__ = ['DAQConfiguration', 'roga_plugndaq', 'umik',
           'default_soundcard', 'configs',
           'findDAQDevice']


@dataclass
class DAQConfiguration:
    """
    Initialize a device descriptor

    Args:
        name: Name of the device to appear to the user
        cardname: ALSA name identifier
        cardlongnamematch: Long name according to ALSA
        device_name: ASCII name with which to open the device when connected
        en_format: index in the list of sample formats
        en_input_rate: index of enabled input sampling frequency [Hz]
                            in the list of frequencies.
        en_input_channels: list of channel indices which are used to
                                acquire data from.
        input_sensitivity: List of sensitivity values, in units of [Pa^-1]
        input_gain_settings: If a DAQ supports it, list of indices which
                            corresponds to a position in the possible input
                            gains for each channel. Should only be not equal
                            to None when the hardware supports changing the
                            input gain.
        en_output_rate: index in the list of possible output sampling
                             frequencies.
        en_output_channels: list of enabled output channels


    """

    name: str
    cardname: str
    cardlongnamematch: str
    device_name: str
    en_format: int
    en_input_rate: int
    en_input_channels: list
    input_sensitivity: list
    input_gain_settings: list
    en_input_gain_settings: list = field(default_factory=list)
    en_output_rate: int = -1
    en_output_channels: list = field(default_factory=list)

    def match(self, device):
        """
        Returns True when a device specification dictionary matches to the
        configuration.

        Args:
            device: dictionary specifying device settings
        """
        match = True
        if self.cardlongnamematch is not None:
            match &= self.cardlongnamematch in device.cardlongname
        if self.cardname is not None:
            match &= self.cardname == device.cardname
        if self.device_name is not None:
            match &= self.device_name == device.device_name
        match &= self.en_format < len(device.available_formats)
        match &= self.en_input_rate < len(device.available_input_rates)
        match &= max(
            self.en_input_channels) < device.max_input_channels
        if len(self.en_output_channels) > 0:
            match &= max(
                self.en_output_channels) < device.max_output_channels
        match &= self.en_output_rate < len(
            device.available_output_rates)

        return match


roga_plugndaq = DAQConfiguration(name='Roga-instruments Plug.n.DAQ USB',
                                 cardname='USB Audio CODEC',
                                 cardlongnamematch='Burr-Brown from TI USB'
                                 ' Audio CODEC',
                                 device_name='iec958:CARD=CODEC,DEV=0',
                                 en_format=0,
                                 en_input_rate=2,
                                 en_input_channels=[0],
                                 input_sensitivity=[46.92e-3, 46.92e-3],
                                 input_gain_settings=[-20, 0, 20],
                                 en_input_gain_settings=[1, 1],
                                 en_output_rate=1,
                                 en_output_channels=[False, False]
                                 )
umik = DAQConfiguration(name='UMIK-1',
                        cardname='Umik-1  Gain: 18dB',
                        cardlongnamematch='miniDSP Umik-1  Gain: 18dB',
                        device_name='iec958:CARD=U18dB,DEV=0',
                        en_format=0,
                        en_input_rate=0,
                        en_input_channels=[0],
                        input_sensitivity=[1., 1.],
                        input_gain_settings=[0., 0.],
                        en_input_gain_settings=[0, 0],
                        en_output_rate=0,
                        en_output_channels=[True, True]
                        )

default_soundcard = DAQConfiguration(name="Default device",
                                     cardname=None,
                                     cardlongnamematch=None,
                                     device_name='default',
                                     en_format=0,
                                     en_input_rate=2,
                                     en_input_channels=[0],
                                     input_sensitivity=[1.0, 1.0],
                                     input_gain_settings=[0],
                                     en_input_gain_settings=[0, 0],
                                     en_output_rate=1,
                                     en_output_channels=[]
                                     )
configs = (roga_plugndaq, default_soundcard)


def findDAQDevice(config: DAQConfiguration) -> DeviceInfo:
    """
    Search for a DaQ device for the given configuration.

    Args:
        config: configuration to search a device for
    """
    devices = query_devices()

    for device in devices:
        if config.match(device):
            return device
    return None
